using DataSetRelationsForeignKeyUniqueKey.Common;
using DataSetRelationsForeignKeyUniqueKey.PashaCode;
using Microsoft.Data.SqlClient;
using System.Data;
using System.Linq;


//===========================================================
//        :
// https://www.interestprograms.ru
//     
//===========================================================

namespace DataSetRelationsForeignKeyUniqueKey
{
    public partial class Form1 : Form
    {
        private bool deleteRows = false;

        public Form1()
        {
            InitializeComponent();

            //      .
            DataSet ds = ConnectDatabase.GetData(
                "SELECT * FROM " + Constants.ParentTableName + " ORDER BY Id ASC;" +
                "SELECT * FROM " + Constants.ChildTableName + " ORDER BY Id ASC;");

            //     ,
            //        .
            SetRelationConstraint(ds);

            //       DataGridView.
            BindDataSource(ds, dgvParent, dgvDetail, dgvChild);
        }


        #region   DataGridView

        private void OnDataError(object sender, DataGridViewDataErrorEventArgs e)
        {
            DataGridView dataGridView = (DataGridView)sender;

            MessageBox.Show(" " + dataGridView.DataMember + "\n:  " +
                e.RowIndex + "x" + e.ColumnIndex + "\n" +
                e.Exception.Message, "", MessageBoxButtons.OK, MessageBoxIcon.Warning);
        }

        private void OnDataBindingComplete(object sender, DataGridViewBindingCompleteEventArgs e)
        {
            DataGridView dgv = (DataGridView)sender;

            foreach (DataGridViewRow row in dgv.Rows)
            {
                //       ,     .
                if (row.IsNewRow == false)
                    row.HeaderCell.Value = (row.Index + 1).ToString();
            }
        }

        private void OnCellValueChanged(object sender, DataGridViewCellEventArgs e)
        {
            if (e.ColumnIndex >= 0)
            {
                DataGridView dgv = (DataGridView)sender;

                //      ( )  DataGridView.
                if (dgv.Rows[e.RowIndex].IsNewRow == false)
                {
                    //       .
                    dgv.Rows[e.RowIndex].Cells[e.ColumnIndex].ErrorText = "  !";
                }
            }
        }

        private void OnRowValidated(object sender, DataGridViewCellEventArgs e)
        {
            DataGridView dgv = (DataGridView)sender;
            BindingSource bs = (BindingSource)dgv.DataSource;
            DataSet ds = (DataSet)bs.DataSource;
            string tablename = bs.DataMember;

            SaveModifications(ds, tablename);

            //        .
            foreach (DataGridViewCell cell in dgv.Rows[e.RowIndex].Cells)
            {
                cell.ErrorText = null;
            }
        }

        private void OnKeyDown(object sender, KeyEventArgs e)
        {
            DataGridView dgv = (DataGridView)sender;
            BindingSource bs = (BindingSource)dgv.DataSource;
            DataSet ds = (DataSet)bs.DataSource;
            string tablename = bs.DataMember;

            if (e.KeyCode == Keys.Delete)
            {
                if (ds != null)
                {
                    int numberrowsdelete = dgv.SelectedRows.Count;
                    DataTable? dt = ds.Tables[tablename];

                    if (dt != null)
                    {
                        deleteRows = false;

                        if (MessageBox.Show(
                            " ?\n - " + numberrowsdelete,
                            " !",
                            MessageBoxButtons.YesNo, MessageBoxIcon.Exclamation) == DialogResult.Yes)
                        {

                            deleteRows = true;
                        }
                    }
                }
            }


            //    .
            if (e.KeyCode == Keys.Escape)
            {
                dgv.Sort(dgv.Columns["Id"], System.ComponentModel.ListSortDirection.Ascending);
            }
        }

        private void OnUserDeletingRow(object sender, DataGridViewRowCancelEventArgs e)
        {
            e.Cancel = !deleteRows;
        }

        private void OnRowsRemoved(object sender, DataGridViewRowsRemovedEventArgs e)
        {
            DataGridView dgv = (DataGridView)sender;
            BindingSource bs = (BindingSource)dgv.DataSource;
            DataSet ds = (DataSet)bs.DataSource;
            string tablename = bs.DataMember;
            DataTable? dt = ds.Tables[tablename];

            if (dt != null)
            {
                DeleteRows(dt);
            }
        }

        private void OnRowEnter(object sender, DataGridViewCellEventArgs e)
        {
            DataGridView dgv = (DataGridView)sender;

            if (dgv.Rows[e.RowIndex].IsNewRow == true)
            {
                dgv.Rows[e.RowIndex].Cells["Id"].Value = "-1";
            }
        }

        #endregion


        #region Functions

        void SaveModifications(DataSet ds, string tablename)
        {
            DataTable? dt = ds.Tables[tablename];

            if (dt != null && ds.HasChanges() == true)
            {
                if (ds.HasChanges(DataRowState.Added) == true) InsertRow(dt);

                if (ds.HasChanges(DataRowState.Modified) == true) UpdateCell(dt);

            }
        }

        void InsertRow(DataTable dt)
        {
            string tablename = dt.TableName;

            DataRow? row = dt.AsEnumerable().FirstOrDefault(r => r.RowState == DataRowState.Added);
            if (row == null) return;

            List<SqlParameter> parameters = new();
            List<string> listnames = new();
            List<string> listvalues = new();

            int count = 0;
            foreach (DataColumn col in row.Table.Columns)
            {
                if (col.AutoIncrement == false)
                {
                    listnames.Add("[" + col.ColumnName + "]");

                    SqlParameter parUri = new()
                    {
                        Direction = ParameterDirection.Input,
                        ParameterName = "Param" + count,
                        Value = row[col.ColumnName]
                    };
                    parameters.Add(parUri);
                    count++;

                    listvalues.Add("@" + parUri.ParameterName);
                }
            }

            string names = string.Join(',', listnames);
            string values = string.Join(',', listvalues);

            SqlParameter lastId = new()
            {
                Direction = ParameterDirection.Output,
                ParameterName = "Identity",
                Size = 8
            };
            parameters.Add(lastId);


            string queryString = "INSERT INTO " + tablename + " (" + names + ") VALUES(" + values + ");";
            queryString += "SET @Identity = IDENT_CURRENT('" + tablename + "');";
            //queryString += "SELECT @Identity=IDENT_CURRENT ('Product');";
            ConnectDatabase.SaveData(dt, queryString, parameters);



            dt.Columns["Id"]!.ReadOnly = false;
            row["Id"] = lastId.Value;
            dt.Columns["Id"]!.ReadOnly = true;

            row.AcceptChanges();
        }

        void UpdateCell(DataTable dt)
        {
            string tablename = dt.TableName;

            DataRow? row = dt.AsEnumerable().FirstOrDefault(r => r.RowState == DataRowState.Modified);
            if (row == null) return;

            List<SqlParameter> parameters = new();
            List<string> listnamesvalues = new();
            object id = row["Id"];
            int count = 0;
            foreach (DataColumn col in row.Table.Columns)
            {
                string columnname = col.ColumnName;
                object value = row[col.ColumnName];

                if (col.AutoIncrement == false)
                {
                    SqlParameter parUri = new()
                    {
                        Direction = ParameterDirection.Input,
                        ParameterName = "Param" + count,
                        Value = row[col.ColumnName]
                    };
                    parameters.Add(parUri);
                    count++;

                    listnamesvalues.Add("[" + col.ColumnName + "]=@" + parUri.ParameterName);
                }
            }

            SqlParameter Id = new()
            {
                Direction = ParameterDirection.Input,
                ParameterName = "Identity",
                Value = id
            };
            parameters.Add(Id);

            string namesvalues = string.Join(',', listnamesvalues);
            string queryString = "UPDATE " + tablename + " SET " + namesvalues + " WHERE Id=@Identity";
            ConnectDatabase.SaveData(dt, queryString, parameters);

        }

        void DeleteRows(DataTable dt)
        {
            string tablename = dt.TableName;

            // 2
            DataRow? row = dt.AsEnumerable().FirstOrDefault(r => r.RowState == DataRowState.Deleted);
            if (row != null)
            {
                string? id = row[0, DataRowVersion.Original].ToString();
                if (id == null) return;

                string queryString = "DELETE FROM " + tablename + " WHERE Id=" + id + ";";
                ConnectDatabase.SaveData(dt, queryString);
            }
        }

        void SetRelationConstraint(DataSet ds)
        {
            DataTable parentTable = ds.Tables[Constants.ParentTableName];
            DataTable childTable = ds.Tables[Constants.ChildTableName];

            DataColumn parentColumn = parentTable.Columns["Id"];
            DataColumn childColumn = childTable.Columns["TypeId"];
            DataRelation relation = new("FK_Product_Type", parentColumn, childColumn, true);
            ds.Relations.Add(relation);

            ForeignKeyConstraint foreignKeyConstraint = (ForeignKeyConstraint)childTable.Constraints["FK_Product_Type"];
            foreignKeyConstraint.DeleteRule = Rule.None;

            UniqueConstraint uniqueName = new(parentTable.Columns["Name"]);
            parentTable.Constraints.Add(uniqueName);
            UniqueConstraint uniqueSKU = new(childTable.Columns["SKU"]);
            childTable.Constraints.Add(uniqueSKU);

            UniqueConstraint uniqueTwoColumns = new(new DataColumn[]{ childTable.Columns["TypeId"], childTable.Columns["SKUType"] });
            childTable.Constraints.Add(uniqueTwoColumns);


            //    .
            //   DataSet   DataTable    .
            DataColumn columnId = parentColumn;
            columnId.AutoIncrement = true;
            columnId.ReadOnly = true;
            DataColumn columnId2 = childTable.Columns["Id"];
            columnId2.AutoIncrement = true;
            columnId2.ReadOnly = true;

            //    NULL.
            //   DataSet   DataTable    .
            parentTable.Columns["Name"].AllowDBNull = false;
            childTable.Columns["SKU"].AllowDBNull = false;
        }

        void BindDataSource(DataSet ds, DataGridView dgvParent, DataGridView dgvDetail, DataGridView dgvChild)
        {
            // ===    ===
            BindingSource parent = new();
            BindingSource detail = new();
            BindingSource child = new();

            //   ,
            //    Product.
            parent.DataSource = ds;
            parent.DataMember = Constants.ParentTableName;
            groupBoxParent.Text = " \"" + Constants.ParentTableName + "\"";

            //     .
            detail.DataSource = parent;
            detail.DataMember = "FK_Product_Type";
            groupBoxDetail.Text = "  ";

            //   Product
            child.DataSource = ds;
            child.DataMember = Constants.ChildTableName;

            //  .
            dgvParent.DataSource = parent;
            dgvDetail.DataSource = detail;
            dgvChild.DataSource = child;

            // ===  Type ===
            dgvParent.Columns["Id"].Visible = false;

            // ===

            //==========     ========================

            dgvDetail.Columns["TypeId"].Visible = false;
            DataGridViewComboBoxColumn colbox2 = new();
            colbox2.Name = "TypeName";
            colbox2.DataSource = parent;
            colbox2.DisplayMember = "Name";
            colbox2.ValueMember = "Id";
            colbox2.DataPropertyName = "TypeId";
            colbox2.DisplayIndex = 1;
            colbox2.FlatStyle = FlatStyle.Flat;
            colbox2.DisplayStyle = DataGridViewComboBoxDisplayStyle.Nothing;
            dgvDetail.Columns.Add(colbox2);
            // ================================================


            // ===============  Product ===============

            dgvChild.Columns["Id"].Visible = false;
            groupBoxChild.Text = " \"" + Constants.ChildTableName + "\"";
            dgvChild.Columns["TypeId"].Visible = false;

            DataGridViewComboBoxColumn colbox = new();
            colbox.Name = "TypeName";
            colbox.DataSource = parent;
            colbox.DisplayMember = "Name";
            colbox.ValueMember = "Id";
            colbox.DataPropertyName = "TypeId";
            colbox.DisplayIndex = 1;
            colbox.FlatStyle = FlatStyle.Flat;
            colbox.SortMode = DataGridViewColumnSortMode.Automatic;
            colbox.DisplayStyle = DataGridViewComboBoxDisplayStyle.Nothing;
            dgvChild.Columns.Add(colbox);
            // ==================================
        }

        #endregion

    }
}